@@ -19,8 +19,6 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
-#include <libintl.h>
 #include <string.h>
 #include "window.h"
 #include "xutils.h"
@@ -28,7 +26,6 @@
 #include "wnck-enum-types.h"
 #include "wnck-marshal.h"
 
-#define _(x) dgettext (GETTEXT_PACKAGE, x)
 #define FALLBACK_NAME _("untitled window")
 #define ALL_WORKSPACES (0xFFFFFFFF)
 
@@ -59,33 +56,17 @@ struct _WnckWindowPrivate
   int pid;
   int workspace;
 
-  int preferred_icon_width;
-  int preferred_icon_height;
-  int preferred_mini_icon_width;
-  int preferred_mini_icon_height;
-
   GdkPixbuf *icon;
   GdkPixbuf *mini_icon;
 
+  WnckIconCache *icon_cache;
+  
   WnckWindowActions actions;
 
   int x;
   int y;
   int width;
   int height;
-  
-  /* window has no icon, don't bother querying the
-   * properties again
-   */
-  guint icon_fetch_failed : 1;
-
-  /* whether to make up a standard default icon if none is set on the
-   * window
-   */
-  guint want_fallback_icon : 1;
-
-  /* Icon is a fallback icon */
-  guint icon_is_fallback : 1;
 
   /* window state */
   guint is_minimized : 1;
@@ -145,7 +126,6 @@ static void update_actions   (WnckWindow *window);
 static void unqueue_update   (WnckWindow *window);
 static void queue_update     (WnckWindow *window);
 static void force_update_now (WnckWindow *window);
-static void invalidate_icon  (WnckWindow *window);
 
 static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -189,11 +169,7 @@ wnck_window_init (WnckWindow *window)
   window->priv->icon_name = g_strdup (FALLBACK_NAME);
   window->priv->workspace = ALL_WORKSPACES;
 
-  window->priv->preferred_icon_width = DEFAULT_ICON_WIDTH;
-  window->priv->preferred_icon_height = DEFAULT_ICON_HEIGHT;
-  window->priv->preferred_mini_icon_width = DEFAULT_MINI_ICON_WIDTH;
-  window->priv->preferred_mini_icon_height = DEFAULT_MINI_ICON_HEIGHT;
-  window->priv->want_fallback_icon = TRUE;
+  window->priv->icon_cache = _wnck_icon_cache_new ();
 }
 
 static void
@@ -272,7 +248,13 @@ wnck_window_finalize (GObject *object)
 
   unqueue_update (window);
 
-  invalidate_icon (window);
+  if (window->priv->icon)
+    g_object_unref (G_OBJECT (window->priv->icon));
+
+  if (window->priv->mini_icon)
+    g_object_unref (G_OBJECT (window->priv->mini_icon));
+  
+  _wnck_icon_cache_free (window->priv->icon_cache);
 
   g_free (window->priv->name);
   g_free (window->priv->icon_name);
@@ -447,6 +429,14 @@ wnck_window_get_group_leader (WnckWindow *window)
   return window->priv->group_leader;
 }
 
+gulong
+wnck_window_get_xid (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), None);
+
+  return window->priv->xwindow;
+}
+
 /**
  * wnck_window_get_session_id:
  * @window: a #WnckWindow
@@ -856,38 +846,39 @@ wnck_window_is_active (WnckWindow *window)
 static void
 get_icons (WnckWindow *window)
 {
-  if (window->priv->icon != NULL ||
-      window->priv->icon_fetch_failed)
-    return;
+  GdkPixbuf *icon;
+  GdkPixbuf *mini_icon;
 
-  _wnck_read_icons (window->priv->xwindow,
-                    &window->priv->icon,
-                    window->priv->preferred_icon_width,
-                    window->priv->preferred_icon_height,
-                    &window->priv->mini_icon,
-                    window->priv->preferred_mini_icon_width,
-                    window->priv->preferred_mini_icon_height);
+  icon = NULL;
+  mini_icon = NULL;
+  
+  if (_wnck_read_icons (window->priv->xwindow,
+                        window->priv->icon_cache,
+                        &icon,
+                        DEFAULT_ICON_WIDTH, DEFAULT_ICON_HEIGHT,
+                        &mini_icon,
+                        DEFAULT_MINI_ICON_WIDTH,
+                        DEFAULT_MINI_ICON_HEIGHT))
+    {
+      window->priv->need_emit_icon_changed = TRUE;
+      
+      if (icon)
+        g_object_ref (G_OBJECT (icon));
+      if (mini_icon)
+        g_object_ref (G_OBJECT (mini_icon));
+
+      if (window->priv->icon)
+        g_object_unref (G_OBJECT (window->priv->icon));
+
+      if (window->priv->mini_icon)
+        g_object_unref (G_OBJECT (window->priv->mini_icon));
+
+      window->priv->icon = icon;
+      window->priv->mini_icon = mini_icon;
+    }
 
   g_assert ((window->priv->icon && window->priv->mini_icon) ||
             !(window->priv->icon || window->priv->mini_icon));
-
-  if (window->priv->icon == NULL)
-    window->priv->icon_fetch_failed = TRUE;
-
-  if (window->priv->icon == NULL &&
-      window->priv->want_fallback_icon)
-    {
-      _wnck_get_fallback_icons (&window->priv->icon,
-                                window->priv->preferred_icon_width,
-                                window->priv->preferred_icon_height,
-                                &window->priv->mini_icon,
-                                window->priv->preferred_mini_icon_width,
-                                window->priv->preferred_mini_icon_height);
-
-      window->priv->icon_is_fallback = TRUE;
-    }
-
-  window->priv->need_emit_icon_changed = FALSE;
 }
 
 GdkPixbuf*
@@ -896,6 +887,10 @@ wnck_window_get_icon (WnckWindow *window)
   g_return_val_if_fail (WNCK_IS_WINDOW (window), NULL);
 
   get_icons (window);
+  if (window->priv->need_emit_icon_changed)
+    queue_update (window); /* not done in get_icons since we call that from
+                            * the update
+                            */
 
   return window->priv->icon;
 }
@@ -906,86 +901,29 @@ wnck_window_get_mini_icon (WnckWindow *window)
   g_return_val_if_fail (WNCK_IS_WINDOW (window), NULL);
   
   get_icons (window);
-
+  if (window->priv->need_emit_icon_changed)
+    queue_update (window); /* not done in get_icons since we call that from
+                            * the update
+                            */
+  
   return window->priv->mini_icon;
 }
 
-void
-wnck_window_set_create_fallback_icon (WnckWindow *window,
-                                      gboolean    setting)
-{
-  g_return_if_fail (WNCK_IS_WINDOW (window));
-
-  setting = setting != FALSE;
-
-  if (setting != window->priv->want_fallback_icon)
-    {
-      if ((setting && window->priv->icon == NULL) ||
-          (!setting && window->priv->icon_is_fallback))
-        invalidate_icon (window);
-      
-      window->priv->want_fallback_icon = setting;
-    }
-}
-
 /**
- * wnck_window_set_icon_size:
+ * wnck_window_get_icon_is_fallback:
  * @window: a #WnckWindow
- * @width: preferred width or -1 if you want natural size
- * @height: preferred height or -1 for natural size
- *
- * Width/height all icons should be scaled to; libwnck
- * will pick the best available source icon to scale
- * to this size. If you pass -1 for @width or @height the
- * icon's largest natural size as provided by the application
- * will be used; this is useful if you want to modify the
- * image yourself, and don't want to lose information from
- * the image by having libwnck scale it first.
  *
+ * Checks if we are using a default fallback icon because
+ * none was set on the window.
+ * 
+ * Return value: %TRUE if icon is a fallback
  **/
-void
-wnck_window_set_icon_size (WnckWindow *window,
-                           int         width,
-                           int         height)
+gboolean
+wnck_window_get_icon_is_fallback (WnckWindow *window)
 {
-  g_return_if_fail (WNCK_IS_WINDOW (window));
-  g_return_if_fail (width != 0);
-  g_return_if_fail (height != 0);
-  
-  if (width != window->priv->preferred_icon_width ||
-      height != window->priv->preferred_icon_height)
-    {
-      window->priv->preferred_icon_width = width;
-      window->priv->preferred_icon_height = height;
-      invalidate_icon (window);
-    }
-}
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
 
-/**
- * wnck_window_set_mini_icon_size:
- * @window: a #WnckWindow
- * @width: preferred width or -1 for natural
- * @height: preferred height or -1 for natural
- *
- * Like wnck_window_set_icon_size() but for the mini icon.
- *
- **/
-void
-wnck_window_set_mini_icon_size (WnckWindow *window,
-                                int         width,
-                                int         height)
-{
-  g_return_if_fail (WNCK_IS_WINDOW (window));
-  g_return_if_fail (width != 0);
-  g_return_if_fail (height != 0);
-  
-  if (width != window->priv->preferred_mini_icon_width ||
-      height != window->priv->preferred_mini_icon_height)
-    {
-      window->priv->preferred_mini_icon_width = width;
-      window->priv->preferred_mini_icon_height = height;
-      invalidate_icon (window);
-    }
+  return _wnck_icon_cache_get_is_fallback (window->priv->icon_cache);
 }
 
 /**
@@ -1082,6 +1020,27 @@ wnck_window_is_visible_on_workspace (WnckWindow    *window,
   return wnck_window_is_on_workspace (window, workspace);
 }
 
+/**
+ * wnck_window_set_icon_geometry:
+ * @window: a #WnckWindow
+ * @x: x coordinate in pixels
+ * @y: y coordinate in pixels
+ * @width: width in pixels
+ * @height: height in pixels
+ *
+ * Set the icon geometry for a WnckWindow.
+ */
+void
+wnck_window_set_icon_geometry (WnckWindow *window,
+			       int         x,
+			       int         y,
+			       int         width,
+			       int         height)
+{
+	_wnck_set_icon_geometry (window->priv->xwindow,
+				 x, y, width, height);
+}
+
 /**
  * wnck_window_is_on_workspace:
  * @window: a #WnckWindow
@@ -1171,8 +1130,8 @@ _wnck_window_process_property_notify (WnckWindow *window,
            xevent->xproperty.atom ==
            _wnck_atom_get ("WM_NORMAL_HINTS"))
     {
-      invalidate_icon (window);
-      window->priv->need_emit_icon_changed = TRUE;
+      _wnck_icon_cache_property_changed (window->priv->icon_cache,
+                                         xevent->xproperty.atom);
       queue_update (window);
     }
 }
@@ -1274,20 +1233,7 @@ update_name (WnckWindow *window)
 
   window->priv->need_update_name = FALSE;
 
-  if (window->priv->name == NULL)
-    window->priv->name =
-      _wnck_get_utf8_property (window->priv->xwindow,
-                               _wnck_atom_get ("_NET_WM_VISIBLE_NAME"));
-
-  if (window->priv->name == NULL)
-    window->priv->name =
-      _wnck_get_utf8_property (window->priv->xwindow,
-                               _wnck_atom_get ("_NET_WM_NAME"));
-
-  if (window->priv->name == NULL)
-    window->priv->name =
-      _wnck_get_text_property (window->priv->xwindow,
-                               XA_WM_NAME);
+  window->priv->name = _wnck_get_name (window->priv->xwindow);
 
   if (window->priv->name == NULL)
     window->priv->name = g_strdup (FALLBACK_NAME);
@@ -1424,6 +1370,8 @@ force_update_now (WnckWindow *window)
   update_minimized (window);
   update_workspace (window); /* emits signals */
   update_actions (window);
+
+  get_icons (window);
   
   new_state = COMPRESS_STATE (window);
 
@@ -1435,10 +1383,7 @@ force_update_now (WnckWindow *window)
                           window->priv->actions);
   
   if (window->priv->need_emit_icon_changed)
-    {
-      window->priv->need_emit_icon_changed = FALSE;
-      emit_icon_changed (window);
-    }
+    emit_icon_changed (window);
 }
 
 
@@ -1471,19 +1416,6 @@ unqueue_update (WnckWindow *window)
     }
 }
 
-static void
-invalidate_icon (WnckWindow *window)
-{
-  if (window->priv->icon)
-    g_object_unref (G_OBJECT (window->priv->icon));
-  if (window->priv->mini_icon)
-    g_object_unref (G_OBJECT (window->priv->mini_icon));
-  window->priv->icon = NULL;
-  window->priv->mini_icon = NULL;
-  window->priv->icon_fetch_failed = FALSE;
-  window->priv->icon_is_fallback = FALSE;
-}
-
 static void
 emit_name_changed (WnckWindow *window)
 {
@@ -1513,6 +1445,7 @@ emit_workspace_changed (WnckWindow *window)
 static void
 emit_icon_changed (WnckWindow *window)
 {
+  window->priv->need_emit_icon_changed = FALSE;
   g_signal_emit (G_OBJECT (window),
                  signals[ICON_CHANGED],
                  0);
