@@ -37,24 +37,28 @@ _wnck_get_cardinal (Window  xwindow,
   gulong nitems;
   gulong bytes_after;
   gulong *num;
-  int err;
+  int err, result;
 
   *val = 0;
   
   _wnck_error_trap_push ();
   type = None;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      atom,
-                      0, G_MAXLONG,
-		      False, XA_CARDINAL, &type, &format, &nitems,
-		      &bytes_after, (guchar **)&num);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       atom,
+			       0, G_MAXLONG,
+			       False, XA_CARDINAL, &type, &format, &nitems,
+			       &bytes_after, (guchar **)&num);  
   err = _wnck_error_trap_pop ();
-  if (err != Success)
+  if (err != Success ||
+      result != Success)
     return FALSE;
   
   if (type != XA_CARDINAL)
-    return FALSE;
+    {
+      XFree (num);
+      return FALSE;
+    }
 
   *val = *num;
   
@@ -71,7 +75,7 @@ _wnck_get_wm_state (Window  xwindow)
   gulong nitems;
   gulong bytes_after;
   gulong *num;
-  int err;
+  int err, result;
   Atom wm_state;
   int retval;
 
@@ -80,18 +84,22 @@ _wnck_get_wm_state (Window  xwindow)
   
   _wnck_error_trap_push ();
   type = None;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      wm_state,
-                      0, G_MAXLONG,
-		      False, wm_state, &type, &format, &nitems,
-		      &bytes_after, (guchar **)&num);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       wm_state,
+			       0, G_MAXLONG,
+			       False, wm_state, &type, &format, &nitems,
+			       &bytes_after, (guchar **)&num);  
   err = _wnck_error_trap_pop ();
-  if (err != Success)
+  if (err != Success ||
+      result != Success)
     return retval;
   
   if (type != wm_state)
-    return retval;
+    {
+      XFree (num);
+      return retval;
+    }
 
   retval = *num;
   
@@ -110,24 +118,28 @@ _wnck_get_window (Window  xwindow,
   gulong nitems;
   gulong bytes_after;
   Window *w;
-  int err;
+  int err, result;
 
   *val = 0;
   
   _wnck_error_trap_push ();
   type = None;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      atom,
-                      0, G_MAXLONG,
-		      False, XA_WINDOW, &type, &format, &nitems,
-		      &bytes_after, (guchar **)&w);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       atom,
+			       0, G_MAXLONG,
+			       False, XA_WINDOW, &type, &format, &nitems,
+			       &bytes_after, (guchar **)&w);  
   err = _wnck_error_trap_pop ();
-  if (err != Success)
+  if (err != Success ||
+      result != Success)
     return FALSE;
   
   if (type != XA_WINDOW)
-    return FALSE;
+    {
+      XFree (w);
+      return FALSE;
+    }
 
   *val = *w;
   
@@ -201,23 +213,27 @@ _wnck_get_string_property_latin1 (Window  xwindow,
   gulong nitems;
   gulong bytes_after;
   guchar *str;
-  int result;
+  int err, result;
   char *retval;
   
   _wnck_error_trap_push ();
   str = NULL;
-  XGetWindowProperty (gdk_display,
-                      xwindow, atom,
-                      0, G_MAXLONG,
-		      False, XA_STRING, &type, &format, &nitems,
-		      &bytes_after, (guchar **)&str);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow, atom,
+			       0, G_MAXLONG,
+			       False, XA_STRING, &type, &format, &nitems,
+			       &bytes_after, (guchar **)&str);  
 
-  result = _wnck_error_trap_pop ();
-  if (result != Success)
+  err = _wnck_error_trap_pop ();
+  if (err != Success ||
+      result != Success)
     return NULL;
   
   if (type != XA_STRING)
-    return NULL; /* FIXME memory leak? */
+    {
+      XFree (str);
+      return NULL;
+    }
 
   retval = g_strdup (str);
   
@@ -235,7 +251,7 @@ _wnck_get_utf8_property (Window  xwindow,
   gulong nitems;
   gulong bytes_after;
   guchar *val;
-  int err;
+  int err, result;
   char *retval;
   Atom utf8_string;
 
@@ -244,16 +260,17 @@ _wnck_get_utf8_property (Window  xwindow,
   _wnck_error_trap_push ();
   type = None;
   val = NULL;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      atom,
-                      0, G_MAXLONG,
-		      False, utf8_string,
-                      &type, &format, &nitems,
-		      &bytes_after, (guchar **)&val);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       atom,
+			       0, G_MAXLONG,
+			       False, utf8_string,
+			       &type, &format, &nitems,
+			       &bytes_after, (guchar **)&val);  
   err = _wnck_error_trap_pop ();
 
-  if (err != Success)
+  if (err != Success ||
+      result != Success)
     return NULL;
   
   if (type != utf8_string ||
@@ -291,25 +308,29 @@ _wnck_get_window_list (Window   xwindow,
   gulong nitems;
   gulong bytes_after;
   Window *data;
-  int err;
+  int err, result;
 
   *windows = NULL;
   *len = 0;
   
   _wnck_error_trap_push ();
   type = None;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      atom,
-                      0, G_MAXLONG,
-		      False, XA_WINDOW, &type, &format, &nitems,
-		      &bytes_after, (guchar **)&data);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       atom,
+			       0, G_MAXLONG,
+			       False, XA_WINDOW, &type, &format, &nitems,
+			       &bytes_after, (guchar **)&data);  
   err = _wnck_error_trap_pop ();
-  if (err != Success)
+  if (err != Success ||
+      result != Success)
     return FALSE;
   
   if (type != XA_WINDOW)
-    return FALSE; /* FIXME memory leak? */
+    {
+      XFree (data);
+      return FALSE;
+    }
 
   *windows = g_new (Window, nitems);
   memcpy (*windows, data, sizeof (Window) * nitems);
@@ -331,25 +352,29 @@ _wnck_get_atom_list (Window   xwindow,
   gulong nitems;
   gulong bytes_after;
   Atom *data;
-  int err;
+  int err, result;
 
   *atoms = NULL;
   *len = 0;
   
   _wnck_error_trap_push ();
   type = None;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      atom,
-                      0, G_MAXLONG,
-		      False, XA_ATOM, &type, &format, &nitems,
-		      &bytes_after, (guchar **)&data);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       atom,
+			       0, G_MAXLONG,
+			       False, XA_ATOM, &type, &format, &nitems,
+			       &bytes_after, (guchar **)&data);  
   err = _wnck_error_trap_pop ();
-  if (err != Success)
+  if (err != Success ||
+      result != Success)
     return FALSE;
   
   if (type != XA_ATOM)
-    return FALSE; /* FIXME memory leak? */
+    {
+      XFree (data);
+      return FALSE;
+    }
 
   *atoms = g_new (Atom, nitems);
   memcpy (*atoms, data, sizeof (Atom) * nitems);
@@ -371,25 +396,29 @@ _wnck_get_cardinal_list (Window   xwindow,
   gulong nitems;
   gulong bytes_after;
   gulong *nums;
-  int err;
+  int err, result;
 
   *cardinals = NULL;
   *len = 0;
   
   _wnck_error_trap_push ();
   type = None;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      atom,
-                      0, G_MAXLONG,
-		      False, XA_CARDINAL, &type, &format, &nitems,
-		      &bytes_after, (guchar **)&nums);  
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       atom,
+			       0, G_MAXLONG,
+			       False, XA_CARDINAL, &type, &format, &nitems,
+			       &bytes_after, (guchar **)&nums);  
   err = _wnck_error_trap_pop ();
-  if (err != Success)
+  if (err != Success ||
+      result != Success)
     return FALSE;
   
   if (type != XA_CARDINAL)
-    return FALSE; /* FIXME memory leak ? */
+    {
+      XFree (nums);
+      return FALSE;
+    }
 
   *cardinals = g_new (gulong, nitems);
   memcpy (*cardinals, nums, sizeof (gulong) * nitems);
@@ -400,6 +429,91 @@ _wnck_get_cardinal_list (Window   xwindow,
   return TRUE;
 }
 
+char**
+_wnck_get_utf8_list (Window   xwindow,
+                     Atom     atom)
+{
+  Atom type;
+  int format;
+  gulong nitems;
+  gulong bytes_after;
+  char *val;
+  int err, result;
+  Atom utf8_string;
+  char **retval;
+  int i;
+  int n_strings;
+  char *p;
+  
+  utf8_string = _wnck_atom_get ("UTF8_STRING");
+  
+  _wnck_error_trap_push ();
+  type = None;
+  val = NULL;
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       atom,
+			       0, G_MAXLONG,
+			       False, utf8_string,
+			       &type, &format, &nitems,
+			       &bytes_after, (guchar **)&val);  
+  err = _wnck_error_trap_pop ();
+
+  if (err != Success ||
+      result != Success)
+    return NULL;
+  
+  if (type != utf8_string ||
+      format != 8 ||
+      nitems == 0)
+    {
+      if (val)
+        XFree (val);
+      return NULL;
+    }
+
+  /* I'm not sure this is right, but I'm guessing the
+   * property is nul-separated
+   */
+  i = 0;
+  n_strings = 0;
+  while (i < nitems)
+    {
+      if (val[i] == '\0')
+        ++n_strings;
+      ++i;
+    }
+
+  /* we're guaranteed that val has a nul on the end
+   * by XGetWindowProperty
+   */
+  
+  retval = g_new0 (char*, n_strings + 1);
+
+  p = val;
+  i = 0;
+  while (i < n_strings)
+    {
+      if (!g_utf8_validate (p, -1, NULL))
+        {
+          g_warning ("Property %s contained invalid UTF-8\n",
+                     _wnck_atom_name (atom));
+          XFree (val);
+          g_strfreev (retval);
+          return NULL;
+        }
+
+      retval[i] = g_strdup (p);
+      
+      p = p + strlen (p) + 1;
+      ++i;
+    }
+  
+  XFree (val);
+  
+  return retval;
+}
+
 void
 _wnck_error_trap_push (void)
 {
@@ -480,9 +594,14 @@ filter_func (GdkXEvent  *gdkxevent,
         else
           {
             WnckWindow *window;
+            WnckApplication *app;
 
             window = wnck_window_get (xevent->xany.window);
+            app = wnck_application_get (xevent->xany.window);
 
+            if (app)
+              _wnck_application_process_property_notify (app, xevent);
+            
             if (window)
               _wnck_window_process_property_notify (window, xevent);
           }
@@ -738,6 +857,25 @@ _wnck_get_pid (Window xwindow)
     return val;
 }
 
+char*
+_wnck_get_name (Window xwindow)
+{
+  char *name;
+  
+  name = _wnck_get_utf8_property (xwindow,
+                                  _wnck_atom_get ("_NET_WM_VISIBLE_NAME"));
+
+  if (name == NULL)
+    name = _wnck_get_utf8_property (xwindow,
+                                    _wnck_atom_get ("_NET_WM_NAME"));
+
+  if (name == NULL)
+    name = _wnck_get_text_property (xwindow,
+                                    XA_WM_NAME);
+
+  return name;
+}
+
 void
 _wnck_select_input (Window xwindow,
                     int    mask)
@@ -946,34 +1084,34 @@ read_rgb_icon (Window         xwindow,
   int format;
   gulong nitems;
   gulong bytes_after;
-  int result;
-  gulong *data; /* FIXME should be guint? */
+  int result, err;
+  gulong *data;
   gulong *best;
   int w, h;
   gulong *best_mini;
   int mini_w, mini_h;
   
-  if (sizeof (gulong) != 4)
-    g_warning ("%s: Whoops, I think this function may be broken on 64-bit\n",
-               __FUNCTION__);
-  
   _wnck_error_trap_push ();
   type = None;
   data = NULL;
-  XGetWindowProperty (gdk_display,
-                      xwindow,
-                      _wnck_atom_get ("_NET_WM_ICON"),
-		      0, G_MAXLONG,
-		      False, XA_CARDINAL, &type, &format, &nitems,
-		      &bytes_after, ((guchar **)&data));
+  result = XGetWindowProperty (gdk_display,
+			       xwindow,
+			       _wnck_atom_get ("_NET_WM_ICON"),
+			       0, G_MAXLONG,
+			       False, XA_CARDINAL, &type, &format, &nitems,
+			       &bytes_after, ((guchar **)&data));
   
-  result = _wnck_error_trap_pop ();
+  err = _wnck_error_trap_pop ();
   
-  if (result != Success)
+  if (err != Success ||
+      result != Success)
     return FALSE;
 
   if (type != XA_CARDINAL)
-    return FALSE; /* FIXME memory leak? */
+    {
+      XFree (data);
+      return FALSE;
+    }
   
   if (!find_best_size (data, nitems,
                        ideal_width, ideal_height,
@@ -1240,27 +1378,31 @@ get_kwm_win_icon (Window  xwindow,
   gulong nitems;
   gulong bytes_after;
   Pixmap *icons;
-  int result;
+  int err, result;
 
   *pixmap = None;
   *mask = None;
   
   _wnck_error_trap_push ();
   icons = NULL;
-  XGetWindowProperty (gdk_display, xwindow,
-                      _wnck_atom_get ("KWM_WIN_ICON"),
-                      0, G_MAXLONG,
-		      False,
-                      _wnck_atom_get ("KWM_WIN_ICON"),
-                      &type, &format, &nitems,
-		      &bytes_after, (guchar **)&icons);  
+  result = XGetWindowProperty (gdk_display, xwindow,
+			       _wnck_atom_get ("KWM_WIN_ICON"),
+			       0, G_MAXLONG,
+			       False,
+			       _wnck_atom_get ("KWM_WIN_ICON"),
+			       &type, &format, &nitems,
+			       &bytes_after, (guchar **)&icons);  
 
-  result = _wnck_error_trap_pop ();
-  if (result != Success)
+  err = _wnck_error_trap_pop ();
+  if (err != Success ||
+      result != Success)
     return;
   
   if (type != _wnck_atom_get ("KWM_WIN_ICON"))
-    return; /* FIXME mem leak? */
+    {
+      XFree (icons);
+      return;
+    }
   
   *pixmap = icons[0];
   *mask = icons[1];
@@ -1270,14 +1412,169 @@ get_kwm_win_icon (Window  xwindow,
   return;
 }
 
+typedef enum
+{
+  /* These MUST be in ascending order of preference;
+   * i.e. if we get _NET_WM_ICON and already have
+   * WM_NORMAL_HINTS, we prefer _NET_WM_ICON
+   */
+  USING_NO_ICON,
+  USING_FALLBACK_ICON,
+  USING_KWM_WIN_ICON,
+  USING_WM_NORMAL_HINTS,
+  USING_NET_WM_ICON
+} IconOrigin;
+
+struct _WnckIconCache
+{
+  IconOrigin origin;
+  Pixmap prev_pixmap;
+  Pixmap prev_mask;
+  GdkPixbuf *icon;
+  GdkPixbuf *mini_icon;
+  int ideal_width;
+  int ideal_height;
+  int ideal_mini_width;
+  int ideal_mini_height;
+  guint want_fallback : 1;
+  /* TRUE if these props have changed */
+  guint wm_normal_hints_dirty : 1;
+  guint kwm_win_icon_dirty : 1;
+  guint net_wm_icon_dirty : 1;  
+};
+
+WnckIconCache*
+_wnck_icon_cache_new (void)
+{
+  WnckIconCache *icon_cache;
+
+  icon_cache = g_new (WnckIconCache, 1);
+
+  icon_cache->origin = USING_NO_ICON;
+  icon_cache->prev_pixmap = None;
+  icon_cache->icon = NULL;
+  icon_cache->mini_icon = NULL;
+  icon_cache->ideal_width = -1; /* won't be a legit width */
+  icon_cache->ideal_height = -1;
+  icon_cache->ideal_mini_width = -1;
+  icon_cache->ideal_mini_height = -1;
+  icon_cache->want_fallback = TRUE;
+  icon_cache->wm_normal_hints_dirty = TRUE;
+  icon_cache->kwm_win_icon_dirty = TRUE;
+  icon_cache->net_wm_icon_dirty = TRUE;
+  
+  return icon_cache;
+}
+
+static void
+clear_icon_cache (WnckIconCache *icon_cache,
+                  gboolean       dirty_all)
+{
+  if (icon_cache->icon)
+    g_object_unref (G_OBJECT (icon_cache->icon));
+  icon_cache->icon = NULL;
+  
+  if (icon_cache->mini_icon)
+    g_object_unref (G_OBJECT (icon_cache->mini_icon));
+  icon_cache->mini_icon = NULL;
+
+  icon_cache->origin = USING_NO_ICON;
+
+  if (dirty_all)
+    {
+      icon_cache->wm_normal_hints_dirty = TRUE;
+      icon_cache->kwm_win_icon_dirty = TRUE;
+      icon_cache->net_wm_icon_dirty = TRUE;
+    }
+}
+
 void
-_wnck_read_icons (Window      xwindow,
-                  GdkPixbuf **iconp,
-                  int         ideal_width,
-                  int         ideal_height,
-                  GdkPixbuf **mini_iconp,
-                  int         ideal_mini_width,
-                  int         ideal_mini_height)
+_wnck_icon_cache_free (WnckIconCache *icon_cache)
+{
+  clear_icon_cache (icon_cache, FALSE);
+  
+  g_free (icon_cache);
+}
+
+void
+_wnck_icon_cache_property_changed (WnckIconCache *icon_cache,
+                                   Atom           atom)
+{  
+  if (atom == _wnck_atom_get ("_NET_WM_ICON"))
+    icon_cache->net_wm_icon_dirty = TRUE;
+  else if (atom == _wnck_atom_get ("KWM_WIN_ICON"))
+    icon_cache->kwm_win_icon_dirty = TRUE;
+  else if (atom == _wnck_atom_get ("WM_NORMAL_HINTS"))
+    icon_cache->wm_normal_hints_dirty = TRUE;
+}
+
+gboolean
+_wnck_icon_cache_get_icon_invalidated (WnckIconCache *icon_cache)
+{
+  if (icon_cache->origin <= USING_KWM_WIN_ICON &&
+      icon_cache->kwm_win_icon_dirty)
+    return TRUE;
+  else if (icon_cache->origin <= USING_WM_NORMAL_HINTS &&
+           icon_cache->wm_normal_hints_dirty)
+    return TRUE;
+  else if (icon_cache->origin <= USING_NET_WM_ICON &&
+           icon_cache->net_wm_icon_dirty)
+    return TRUE;
+  else if (icon_cache->origin < USING_FALLBACK_ICON &&
+           icon_cache->want_fallback)
+    return TRUE;
+  else if (icon_cache->origin == USING_NO_ICON)
+    return TRUE;
+  else if (icon_cache->origin == USING_FALLBACK_ICON &&
+           !icon_cache->want_fallback)
+    return TRUE;
+  else
+    return FALSE;
+}
+
+void
+_wnck_icon_cache_set_want_fallback (WnckIconCache *icon_cache,
+                                    gboolean       setting)
+{
+  icon_cache->want_fallback = TRUE;
+}
+
+gboolean
+_wnck_icon_cache_get_is_fallback (WnckIconCache *icon_cache)
+{
+  return icon_cache->origin == USING_FALLBACK_ICON;
+}
+
+static void
+replace_cache (WnckIconCache *icon_cache,
+               IconOrigin     origin,
+               GdkPixbuf     *new_icon,
+               GdkPixbuf     *new_mini_icon)
+{
+  clear_icon_cache (icon_cache, FALSE);
+  
+  icon_cache->origin = origin;
+
+  if (new_icon)
+    g_object_ref (G_OBJECT (new_icon));
+
+  icon_cache->icon = new_icon;
+
+  if (new_mini_icon)
+    g_object_ref (G_OBJECT (new_mini_icon));
+
+  icon_cache->mini_icon = new_mini_icon;
+}
+
+gboolean
+_wnck_read_icons (Window         xwindow,
+                  WnckIconCache *icon_cache,
+                  GdkPixbuf    **iconp,
+                  int            ideal_width,
+                  int            ideal_height,
+                  GdkPixbuf    **mini_iconp,
+                  int            ideal_mini_width,
+                  int            ideal_mini_height)
 {
   guchar *pixdata;     
   int w, h;
@@ -1286,68 +1583,171 @@ _wnck_read_icons (Window      xwindow,
   Pixmap pixmap;
   Pixmap mask;
   XWMHints *hints;
+
+  /* Return value is whether the icon changed */
+  
+  g_return_val_if_fail (icon_cache != NULL, FALSE);
   
   *iconp = NULL;
   *mini_iconp = NULL;
+  
+  if (ideal_width != icon_cache->ideal_width ||
+      ideal_height != icon_cache->ideal_height ||
+      ideal_mini_width != icon_cache->ideal_mini_width ||
+      ideal_mini_height != icon_cache->ideal_mini_height)
+    clear_icon_cache (icon_cache, TRUE);
+  
+  icon_cache->ideal_width = ideal_width;
+  icon_cache->ideal_height = ideal_height;
+  icon_cache->ideal_mini_width = ideal_mini_width;
+  icon_cache->ideal_mini_height = ideal_mini_height;
 
+  if (!_wnck_icon_cache_get_icon_invalidated (icon_cache))
+    return FALSE; /* we have no new info to use */
+  
   pixdata = NULL;
-  
-  if (read_rgb_icon (xwindow,
-                     ideal_width, ideal_height,
-                     ideal_mini_width, ideal_mini_height,
-                     &w, &h, &pixdata,
-                     &mini_w, &mini_h, &mini_pixdata))
+
+  /* Our algorithm here assumes that we can't have for example origin
+   * < USING_NET_WM_ICON and icon_cache->net_wm_icon_dirty == FALSE
+   * unless we have tried to read NET_WM_ICON.
+   *
+   * Put another way, if an icon origin is not dirty, then we have
+   * tried to read it at the current size. If it is dirty, then
+   * we haven't done that since the last change.
+   */
+   
+  if (icon_cache->origin <= USING_NET_WM_ICON &&
+      icon_cache->net_wm_icon_dirty)
+
     {
-      *iconp = gdk_pixbuf_new_from_data (pixdata,
-                                         GDK_COLORSPACE_RGB,
-                                         TRUE,
-                                         8,
-                                         w, h, w * 4,
-                                         free_pixels,
-                                         NULL);
+      icon_cache->net_wm_icon_dirty = FALSE;
+      
+      if (read_rgb_icon (xwindow,
+                         ideal_width, ideal_height,
+                         ideal_mini_width, ideal_mini_height,
+                         &w, &h, &pixdata,
+                         &mini_w, &mini_h, &mini_pixdata))
+        {
+          *iconp = gdk_pixbuf_new_from_data (pixdata,
+                                             GDK_COLORSPACE_RGB,
+                                             TRUE,
+                                             8,
+                                             w, h, w * 4,
+                                             free_pixels,
+                                             NULL);
+          
+          *mini_iconp = gdk_pixbuf_new_from_data (mini_pixdata,
+                                                  GDK_COLORSPACE_RGB,
+                                                  TRUE,
+                                                  8,
+                                                  mini_w, mini_h, mini_w * 4,
+                                                  free_pixels,
+                                                  NULL);
 
-      *mini_iconp = gdk_pixbuf_new_from_data (mini_pixdata,
-                                              GDK_COLORSPACE_RGB,
-                                              TRUE,
-                                              8,
-                                              mini_w, mini_h, mini_w * 4,
-                                              free_pixels,
-                                              NULL);
+          replace_cache (icon_cache, USING_NET_WM_ICON,
+                         *iconp, *mini_iconp);
 
-      return;
-    }
-  
-  _wnck_error_trap_push ();
-  hints = XGetWMHints (gdk_display, xwindow);
-  _wnck_error_trap_pop ();
-  pixmap = None;
-  mask = None;
-  if (hints)
-    {
-      if (hints->flags & IconPixmapHint)
-        pixmap = hints->icon_pixmap;
-      if (hints->flags & IconMaskHint)
-        mask = hints->icon_mask;
-
-      XFree (hints);
-      hints = NULL;
+          return TRUE;
+        }
     }
 
-  if (pixmap != None &&
-      try_pixmap_and_mask (pixmap, mask,
-                           iconp, ideal_width, ideal_height,
-                           mini_iconp, ideal_mini_width, ideal_mini_height))
-    return;
+  if (icon_cache->origin <= USING_WM_NORMAL_HINTS &&
+      icon_cache->wm_normal_hints_dirty)
+    {
+      icon_cache->wm_normal_hints_dirty = FALSE;
+      
+      _wnck_error_trap_push ();
+      hints = XGetWMHints (gdk_display, xwindow);
+      _wnck_error_trap_pop ();
+      pixmap = None;
+      mask = None;
+      if (hints)
+        {
+          if (hints->flags & IconPixmapHint)
+            pixmap = hints->icon_pixmap;
+          if (hints->flags & IconMaskHint)
+            mask = hints->icon_mask;
 
-  get_kwm_win_icon (xwindow, &pixmap, &mask);
+          XFree (hints);
+          hints = NULL;
+        }
 
-  if (pixmap != None &&
-      try_pixmap_and_mask (pixmap, mask,
-                           iconp, ideal_width, ideal_height,
-                           mini_iconp, ideal_mini_width, ideal_mini_height))
-    return;
+      /* We won't update if pixmap is unchanged;
+       * avoids a get_from_drawable() on every geometry
+       * hints change
+       */
+      if ((pixmap != icon_cache->prev_pixmap ||
+           mask != icon_cache->prev_mask) &&
+          pixmap != None)
+        {
+          if (try_pixmap_and_mask (pixmap, mask,
+                                   iconp, ideal_width, ideal_height,
+                                   mini_iconp, ideal_mini_width, ideal_mini_height))
+            {
+              icon_cache->prev_pixmap = pixmap;
+              icon_cache->prev_mask = mask;
+
+              replace_cache (icon_cache, USING_WM_NORMAL_HINTS,
+                             *iconp, *mini_iconp);
+
+              return TRUE;
+            }
+        }
+    }
+
+  if (icon_cache->origin <= USING_KWM_WIN_ICON &&
+      icon_cache->kwm_win_icon_dirty)
+    {
+      icon_cache->kwm_win_icon_dirty = FALSE;
+      
+      get_kwm_win_icon (xwindow, &pixmap, &mask);
+
+      if ((pixmap != icon_cache->prev_pixmap ||
+           mask != icon_cache->prev_mask) &&
+          pixmap != None)
+        {
+          if (try_pixmap_and_mask (pixmap, mask,
+                                   iconp, ideal_width, ideal_height,
+                                   mini_iconp, ideal_mini_width, ideal_mini_height))
+            {
+              icon_cache->prev_pixmap = pixmap;
+              icon_cache->prev_mask = mask;
+
+              replace_cache (icon_cache, USING_KWM_WIN_ICON,
+                             *iconp, *mini_iconp);
+              
+              return TRUE;
+            }
+        }
+    }
   
-  /* No further ideas... */
+  if (icon_cache->want_fallback &&
+      icon_cache->origin < USING_FALLBACK_ICON)
+    {
+      _wnck_get_fallback_icons (iconp,
+                                ideal_width,
+                                ideal_height,
+                                mini_iconp,
+                                ideal_mini_width,
+                                ideal_mini_height);
+
+      replace_cache (icon_cache, USING_FALLBACK_ICON,
+                     *iconp, *mini_iconp);
+
+      return TRUE;
+    }
+
+  if (!icon_cache->want_fallback &&
+      icon_cache->origin == USING_FALLBACK_ICON)
+    {
+      /* Get rid of current icon */
+      clear_icon_cache (icon_cache, FALSE);
+
+      return TRUE;
+    }
+
+  /* found nothing new */
+  return FALSE;
 }
 
 #ifdef HAVE_GDK_PIXBUF_NEW_FROM_STREAM
@@ -1465,6 +1865,31 @@ _wnck_get_window_position (Window xwindow,
     *yp = y;
 }
 
+void
+_wnck_set_icon_geometry  (Window xwindow,
+			  int    x,
+			  int    y,
+			  int    width,
+			  int    height)
+{
+  gulong data[4];
+
+  data[0] = x;
+  data[1] = y;
+  data[2] = width;
+  data[3] = height;
+  
+  _wnck_error_trap_push ();
+
+  XChangeProperty (gdk_display,
+		   xwindow,
+		   _wnck_atom_get ("_NET_WM_ICON_GEOMETRY"),
+		   XA_CARDINAL, 32, PropModeReplace,
+		   (guchar *)&data, 4);
+
+  _wnck_error_trap_pop ();
+}
+
 void
 _wnck_set_dock_type_hint (Window xwindow)
 {
